package com.gigaspaces.ucs.deploy;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.util.StringTokenizer;
import java.util.logging.Level;

import org.apache.commons.vfs.FileObject;
import org.apache.commons.vfs.FileSelectInfo;
import org.apache.commons.vfs.FileSelector;
import org.apache.commons.vfs.FileSystemException;
import org.apache.commons.vfs.FileSystemManager;
import org.apache.commons.vfs.FileSystemOptions;
import org.apache.commons.vfs.VFS;
import org.apache.commons.vfs.provider.sftp.SftpFileSystemConfigBuilder;
import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.taskdefs.optional.ssh.SSHExec;
import org.openspaces.admin.Admin;
import org.openspaces.admin.machine.Machine;

/************
 * The agentless installer class is responsible for installing gigaspaces on a
 * remote machine, using only ssh. It will upload all relevant files and start
 * the gigaspaces agent.
 * 
 * File transfer is handled using apache commons vfs.
 * 
 * @author barakme
 * 
 */
public class AgentlessInstaller {

	private static final int LOCAL_FILE_BUFFER_SIZE = 1024;

	/************
	 * Details for an installation request.
	 * 
	 * @author barakme
	 * 
	 */
	public static class InstallationDetails {
		// IP of the machine to install. ssh must already be running.
		// May be null, if MAC based installation is used.
		private String targetIP;

		// ssh username
		private String username;

		// ssh password
		private String password;

		// Locator that gigaspaces agent will use.
		private String locator;

		// An instance of the Gigaspaces Admin API. If passed,
		// will be used to check when an agent joins the cluster.
		private Admin admin;

		// true if this machine should act as the LUS and ESM
		private boolean isLus;

		// directory on local machine where installation files are
		// placed. At a minimum, the start-sm.sh should be placed there.
		// Other files may include the gigaspaces installation, java,
		// and any other required scripts.
		private String localDir;

		// The directory on the remote machine where installation
		// files will be uploaded to.
		private String remoteDir;

		public Admin getAdmin() {
			return admin;
		}

		public String getLocalDir() {
			return localDir;
		}

		public String getLocator() {
			return locator;
		}

		public String getPassword() {
			return password;
		}

		public String getRemoteDir() {
			return remoteDir;
		}

		public String getTargetIP() {
			return targetIP;
		}

		public String getUsername() {
			return username;
		}

		public boolean isLus() {
			return isLus;
		}

		public void setAdmin(final Admin admin) {
			this.admin = admin;
		}

		public void setLocalDir(final String localDir) {
			this.localDir = localDir;
		}

		public void setLocator(final String locator) {
			this.locator = locator;
		}

		public void setLus(final boolean isLus) {
			this.isLus = isLus;
		}

		public void setPassword(final String password) {
			this.password = password;
		}

		public void setRemoteDir(final String remoteDir) {
			this.remoteDir = remoteDir;
		}

		public void setTargetIP(final String targetIP) {
			this.targetIP = targetIP;
		}

		public void setUsername(final String username) {
			this.username = username;
		}
	}

	private static final int MAC_TEST_INTERVAL = 60000;
	private static java.util.logging.Logger logger = java.util.logging.Logger.getLogger(AgentlessInstaller.class
			.getName());
	private static final int IP_LOOKUP_TIMEOUT = 1000 * 60 * 8;
	private static final long SSH_COMMAND_TIMEOUT = 5000;
	private static final int CONNECTION_TEST_INTERVAL_MILLIS = 5000;
	private static final int CONNECTION_TEST_CONNECT_TIMEOUT_MILLIS = 10000;
	private static final int SSH_PORT = 22;
	private static final int SSH_CONNECTION_RETRIES = 5;
	private static final int CLUSTER_JOIN_TIMEOUT_MILLIS = 60000;
	private static final int LUS_PORT = 4164;
	private static final int LUS_CONNECTION_RETRIES = 5;

	/******
	 * .
	 */
	public AgentlessInstaller() {

	}

	/*******
	 * Checks if a tcp connection to a remote machine and port is possible.
	 * 
	 * @param ip
	 *            remote machine ip.
	 * @param port
	 *            remote machine port.
	 * @param retries
	 *            number of retries.
	 * @return true if connection succeeded, false otherwise.
	 */
	public boolean checkConnection(final String ip, final int port, final int retries) {
		for (int i = 0; i < retries; ++i) {
			final Socket sock = new Socket();
			try {
				logger.info("Checking connection to: " + ip + ":" + port);
				sock.connect(new InetSocketAddress(ip, port), CONNECTION_TEST_CONNECT_TIMEOUT_MILLIS);
				sock.close();
				return true;
			} catch (final IOException e) {
				// ignore
			}

			try {
				Thread.sleep(CONNECTION_TEST_INTERVAL_MILLIS);
			} catch (final InterruptedException e) {
				// ignore
			}
		}

		return false;

	}

	/****
	 * Copies files from local dir to remote dir.
	 * 
	 * @param host
	 *            host name or ip address of remote machine.
	 * @param username
	 *            ssh username of remote machine.
	 * @param password
	 *            ssh password of remote machine.
	 * @param srcDir
	 *            local directory.
	 * @param toDir
	 *            remote directory.
	 * @throws IOException
	 *             in case of an error during file transfer.
	 */
	private void copyFiles(final String host, final String username, final String password, final String srcDir,
			final String toDir) throws IOException {

		setUpScriptFiles(new File(srcDir));

		final FileSystemOptions opts = new FileSystemOptions();

		SftpFileSystemConfigBuilder.getInstance().setStrictHostKeyChecking(opts, "no");

		final FileSystemManager mng = VFS.getManager();

		final FileObject localDir = mng.resolveFile("file:" + srcDir);
		final String scpTarget = "sftp://" + username + ":" + password + "@" + host + toDir;
		final FileObject remoteDir = mng.resolveFile(scpTarget, opts);

		logger.info("Copying files to: " + scpTarget + " from local dir: " + localDir.getName().getPath());
		try {
			remoteDir.copyFrom(localDir, new FileSelector() {

				@Override
				public boolean includeFile(final FileSelectInfo fileInfo) throws Exception {
					final FileObject remoteFile = mng.resolveFile(remoteDir, localDir.getName().getRelativeName(
							fileInfo.getFile().getName()));
					if (!remoteFile.exists()) {
						logger.info(fileInfo.getFile().getName().getBaseName() + " missing on server");
						return true;
					}

					final long remoteSize = remoteFile.getContent().getSize();
					final long localSize = fileInfo.getFile().getContent().getSize();
					final boolean res = (localSize != remoteSize);
					if (res) {
						logger.info(fileInfo.getFile().getName().getBaseName() + " different on server");
					}
					return res;

				}

				@Override
				public boolean traverseDescendents(final FileSelectInfo fileInfo) throws Exception {
					return true;
				}
			});
		} finally {

			mng.closeFileSystem(remoteDir.getFileSystem());
		}
	}

	private String createMacString(final String[] macs) {
		final StringBuilder builder = new StringBuilder();
		for (final String mac : macs) {
			builder.append(mac).append(';');
		}
		return builder.toString();
	}

	private Process executeScriptProcess(final String scriptLocation, final String macString, final String myIp)
			throws IOException, InterruptedException {
		logger.info("Executing script: " + scriptLocation + " " + myIp + " " + macString);
		final ProcessBuilder pb = new ProcessBuilder(scriptLocation, myIp, macString);
		final Process p = pb.start();
		p.waitFor();
		return p;
	}

	/******
	 * Performs installation on a remote machine with a known IP.
	 * 
	 * @param details
	 *            the installation details.
	 */
	public void installOnMachineWithIP(final InstallationDetails details) {
		// checking for SSH connection
		logger.info("Checking connection to: " + details.getTargetIP() + ":" + SSH_PORT);
		if (!checkConnection(details.getTargetIP(), SSH_PORT, SSH_CONNECTION_RETRIES)) {
			logger.severe("Failed to open a connection to the target machine (" + details.getTargetIP() + ") "
					+ "on port " + SSH_PORT + ". Please check firewall settings");
			throw new IllegalStateException("Failed to open a connection to the target machine ("
					+ details.getTargetIP() + ") " + "on port " + SSH_PORT + ". Please check firewall settings");
		}

		// upload script file
		try {
			copyFiles(details.getTargetIP(), details.getUsername(), details.getPassword(), details.getLocalDir(),
					details.getRemoteDir());
		} catch (final FileSystemException e) {
			logger.log(Level.SEVERE, "Copy of files to remote server failed.", e);
			throw new IllegalStateException("Failed to upload files to server", e);
		} catch (final IOException e) {
			logger.log(Level.SEVERE, "Copy of files to remote server failed.", e);
			throw new IllegalStateException("Failed to upload files to server", e);
		}

		// SSH is ready, call startup script.
		logger.info("Calling startup script on target: " + details.getTargetIP() + " with LOCATOR="
				+ details.getLocator());

		String cmd = null;
		if (details.isLus()) {
			cmd = "chmod +x /opt/gs-files/start-esm.sh;" + "/opt/gs-files/start-esm.sh " + details.getLocator()
					+ " lus  &";
		} else {
			cmd = "chmod +x /opt/gs-files/start-esm.sh;" + "/opt/gs-files/start-esm.sh " + details.getLocator()
					+ " agent &";
		}

		sshCommand(details.getTargetIP(), cmd, details.getUsername(), details.getPassword());

		if (details.isLus()) {
			// Wait for LUS to become available
			if (!checkConnection(details.getTargetIP(), LUS_PORT, LUS_CONNECTION_RETRIES)) {
				logger.severe("Failed to open a connection to the target machine (" + details.getTargetIP() + ") "
						+ "on port " + LUS_PORT + ". Please check firewall settings");

				throw new IllegalStateException("Failed to open a connection to the target machine ("
						+ details.getTargetIP() + ") " + "on port " + LUS_PORT + ". Please check firewall settings");

			}
		} else {
			// wait for the server to join
			if (details.getAdmin() != null) {
				waitForServerToJoin(details.getTargetIP(), details.getAdmin());
			}
		}
		// At this point, the server is ready to be used.

	}

	/*********
	 * Performs installation on a remote machine for which we do not have a
	 * hostname or IP address, only a MAC address. The installer will attempt to
	 * discover the IP address of the machine by performing several rounds of
	 * ping sweeps on the local network. This may take a while.
	 * 
	 * @param mac
	 *            mac address of remote machine.
	 * @param myIp
	 *            IP of this machine. Will be used as base for ping sweep.
	 * @param details
	 *            the installation details.
	 */
	public void installOnMachineWithMAC(final String mac, final String myIp, final InstallationDetails details) {
		installOnMachineWithMAC(new String[] { mac }, myIp, details);
	}

	/*********
	 * Performs installation on a remote machine for which we do not have a
	 * hostname or IP address, only a MAC address. The installer will attempt to
	 * discover the IP address of the machine by performing several rounds of
	 * ping sweeps on the local network. This may take a while.
	 * 
	 * @param macs
	 *            mac addresses of remote machine.
	 * @param myIp
	 *            IP of this machine. Will be used as base for ping sweep.
	 * @param details
	 *            the installation details.
	 */
	public void installOnMachineWithMAC(final String[] macs, final String myIp, final InstallationDetails details) {
		final long startTime = System.currentTimeMillis();
		final long endTime = startTime + IP_LOOKUP_TIMEOUT;
		final String macString = createMacString(macs);

		String cmd = details.getLocalDir();
		if (!cmd.endsWith("/")) {
			cmd += "/";
		}
		cmd += "ipFromMac.sh";

		try {

			while (System.currentTimeMillis() < endTime) {
				final Process proc = executeScriptProcess(cmd, macString, myIp);

				final String output = readProcessResult(proc);
				logger.info("Script output was: " + output);

				final String ip = parseOutput(output);

				if (ip != null) { // found the IP!
					logger.info("IP Discovered: " + ip);
					details.setTargetIP(ip);
					logger.info("Installing on IP: " + ip);
					installOnMachineWithIP(details);
					return;
				} else {
					logger.info("IP was not discovered yet, sleeping");
				}

				try {
					Thread.sleep(MAC_TEST_INTERVAL);
				} catch (final InterruptedException ie) {
					// ignore
				}
			}
			logger.severe("Could not find the IP address of MAC: " + macString
					+ " within the timeout. The target server networking may not have started properly.");

		} catch (final IOException e) {
			logger.log(Level.SEVERE, e.getMessage(), e);
		} catch (final InterruptedException e) {
			logger.log(Level.SEVERE, e.getMessage(), e);
		}

	}

	private String parseOutput(final String output) {
		final StringTokenizer tokenizer = new StringTokenizer(output, ";\n");
		String ip = null;
		while (tokenizer.hasMoreTokens()) {
			final String temp = tokenizer.nextToken().trim();
			if (temp.length() > 0) {
				if (ip == null) {
					ip = temp;
				} else {
					if (!ip.equals(temp)) {
						throw new IllegalArgumentException("Founf two IPS that match the given set of MAC Addresses - "
								+ ip + " and " + temp);
					}
				}
			}

		}
		return ip;
	}

	private String readProcessResult(final Process p) throws IOException {
		BufferedReader reader = null;
		try {
			reader = new BufferedReader(new InputStreamReader(p.getInputStream()));
			final StringBuilder builder = new StringBuilder();

			while (true) {
				final String line = reader.readLine();
				if (line == null) {
					break;
				}
				builder.append(line);
			}

			final String output = builder.toString();
			return output;
		} finally {
			if (reader != null) {
				reader.close();
			}

		}
	}

	private void setUpFile(final String fileName, final String resourcePrefix, final File localBaseDir)
			throws IOException {

		final File file = new File(localBaseDir.getAbsolutePath(), fileName);
		if (file.exists()) {
			return;
		}

		final InputStream input = Thread.currentThread().getContextClassLoader().getResourceAsStream(
				resourcePrefix + fileName);
		if (input == null) {
			throw new IllegalArgumentException(
					"Could not find resource: "
							+ resourcePrefix
							+ fileName
							+ ". Make sure that the required jar files are on your classpath. " +
									"If using a debugger, add the project root directory to your classpath.");
		}

		FileOutputStream output = null;

		try {
			output = new FileOutputStream(file);

			final byte[] buf = new byte[LOCAL_FILE_BUFFER_SIZE];
			int len;
			while ((len = input.read(buf)) > 0) {
				output.write(buf, 0, len);
			}
		} finally {
			if (input != null) {
				input.close();
			}
			if (output != null) {
				output.close();
			}
		}
	}

	/******
	 * Checks if these default files already exist in the local dir. If they do
	 * not, copy the default files from the jar file.
	 * 
	 * @param localBaseDir
	 *            local installation files dir.
	 * @throws IOException .
	 */
	private void setUpScriptFiles(final File localBaseDir) throws IOException {
		setUpFile("ipFromMac.sh", "config/", localBaseDir);
		setUpFile("ipFromMac.xslt", "config/", localBaseDir);
		setUpFile("start-esm.sh", "config/", localBaseDir);

	}

	private void sshCommand(final String host, final String command, final String username, final String password) {

		final SSHExec task = new SSHExec();
		task.setCommand(command);
		task.setHost(host);
		task.setTrust(true);
		task.setUsername(username);
		task.setPassword(password);
		task.setTimeout(SSH_COMMAND_TIMEOUT);

		try {
			logger.info("Executing command: " + command + " on " + host);
			task.execute();
		} catch (final BuildException e) {
			// There really should be a better way to check that this is
			// a timeout
			if ((e.getMessage() != null) && e.getMessage().contains("Timeout")) {
				logger.info("ssh execution ends with timeout: " + e.getCause());
			} else {
				throw new IllegalStateException("Command " + command + " failed to execute: " + e.getMessage(), e);
			}

		}
	}

	private void waitForServerToJoin(final String address, final Admin admin) {

		final long startTime = System.currentTimeMillis();
		while (true) {
			logger.fine("Checking if " + address + " has joined");
			final Machine machine = admin.getMachines().getMachineByHostAddress(address);
			if (machine != null) {
				logger.fine(address + " has joined the cluster!");
				return;
			}
			logger.fine(address + " has not joined yet, sleeping");
			final long now = System.currentTimeMillis();
			if (now - startTime > CLUSTER_JOIN_TIMEOUT_MILLIS) {
				logger.severe("Server " + address + " failed to join cluster before timeout expired");
				throw new IllegalStateException("Server " + address + " failed to join cluster before timeout expired");
			}
		}

	}
}
