package com.gigaspaces.rackspace.examples;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Reader;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.TimeUnit;

import org.jclouds.rackspace.cloudservers.domain.Server;
import org.openspaces.admin.Admin;
import org.openspaces.admin.AdminFactory;
import org.openspaces.admin.esm.ElasticServiceManager;
import org.openspaces.admin.esm.deployment.ElasticDataGridDeployment;
import org.openspaces.admin.pu.ProcessingUnit;
import org.openspaces.admin.space.Space;
import org.openspaces.core.GigaSpace;
import org.openspaces.grid.esm.ElasticScaleHandlerConfig;

import com.gigaspaces.rackspace.Deployer;
import com.j_spaces.core.client.SQLQuery;

/*******
 * A demonstration of the ESM managed data grid functionality.
 * 
 * @author barakme
 * 
 */
public final class ESMDemo {

	private static final String RACKSPACE_PROPERTIES_FILE = "rackspace.properties";

	/****
	 * Tester for the Gigaspaces JDBC functionality.
	 */
	public static final class JDBCQuery {
		private JDBCQuery() {

		}

		/****
		 * Runs the JDBC test.
		 * 
		 * @param locator
		 *            Locator for the Lookup Service.
		 * @throws Exception
		 *             if the test failed.
		 */
		public static void testJDBC(final String locator) throws Exception {

			Connection conn = null;
			Statement st = null;
			try {
				Class.forName("com.j_spaces.jdbc.driver.GDriver").newInstance();
				final String url = "jdbc:gigaspaces:url:jini://*/*/myElasticDataGrid?locators=" + locator;
				conn = DriverManager.getConnection(url);
				st = conn.createStatement();
				final String query = "SELECT * FROM com.gigaspaces.rackspace.examples.Data";
				final ResultSet rs = st.executeQuery(query);

				// Iterate through the result set
				int i = 0;
				while (rs.next()) {

					System.out.println("Data [" + (i++) + "] " + rs.getString("data"));// +

				}
			} finally {
				if (st != null) {
					st.close();
				}
				if (conn != null) {
					conn.close();
				}
			}

		}

	}

	/********
	 * Create the test data.
	 * 
	 * @param i
	 *            index.
	 * @return the test data.
	 */
	public static Map<String, Object> createInfo(final long i) {
		final Map<String, Object> info = new HashMap<String, Object>();
		info.put("address", i + " Broadway");
		info.put("socialSecurity", 1232287642L + i);
		info.put("salary", 10000 + i);
		return info;
	}

	/**
	 * The main method for the demo.
	 * 
	 * @param args
	 *            not used.
	 * @throws Exception .
	 */
	public static void main(final String[] args) throws Exception {

		final String gridName = "myElasticDataGrid";
		Properties props = loadProperties();

		final String user = props.getProperty("user");
		final String apiKey = props.getProperty("apiKey");

		// Creating a data grid on demand
		System.out.println("Deploying Data Grid..");

		// Launch the head node, with the GigaSpaces Elastic Scaling Module
		final String locator = ESMDemo.startESMNode("gs-esm-manager", user, apiKey);

		// Starts the GS-UI with the locator
		startGUI(locator);

		// Initialize the GigaSpaces Admin API with the locator for the head
		// node
		final Admin admin = new AdminFactory().addLocator(locator).createAdmin();

		// Use the Admin API to deploy an elastic grid
		final ElasticServiceManager esm = admin.getElasticServiceManagers().waitForAtLeastOne();
		final ProcessingUnit pu = esm.deploy(new ElasticDataGridDeployment(gridName).elasticScaleHandler(
				new ElasticScaleHandlerConfig("com.gigaspaces.rackspace.esm.RackspaceElasticHandler").addProperty(
						"user", user).addProperty("apiKei", apiKey).addProperty("machineNamePrefix", "gs-esm-gsa")
						.addProperty("gridName", gridName)).capacity("1G", "2G").highlyAvailable(false)
				.publicDeploymentIsolation().maximumJavaHeapSize("250m").initialJavaHeapSize("250m"));

		try {

			System.out.println("Waiting for data grid deployment completion ..");
			final Space space = pu.waitForSpace();

			while (!space.waitFor(space.getTotalNumberOfInstances(), 10, TimeUnit.SECONDS)) {
				System.out.println("Waiting for all partitions to deploy. " + "Available: "
						+ space.getNumberOfInstances() + ", " + "Required: " + space.getTotalNumberOfInstances());
			}

			System.out.println("All instances have been deployed, the Data Grid is ready for action");

			// Write some data...
			final GigaSpace gigaSpace = space.getGigaSpace();

			System.out.println("Writing 1000 objects");
			for (long i = 0; i < 1000; i++) {
				gigaSpace.write(new Data(i, "message" + i, ESMDemo.createInfo(i)));
			}

			System.out.println("Reading 1000 objects");
			final Data[] d = gigaSpace.readMultiple(new SQLQuery<Data>(Data.class,
					"info.salary < 11000 and info.salary >= 10000"), Integer.MAX_VALUE);

			System.out.println("SQLQuery<Data>(Data.class, info.salary < 15000 and info.salary >= 8000) results -> ["
					+ d.length + " ]");

			System.out.println("Reading 1000 using JDBC connector objects");

			JDBCQuery.testJDBC(locator);
			// System.out.println("Clearing...");
			// pu.undeploy();
			System.out.println("Done!");
			System.exit(0);

		} catch (final Exception ex) {
			// for debugging purposes clean the deployment in case of exception
			ex.printStackTrace();
			System.out.println("Caught exception: Undeploying... ");
			pu.undeploy();
			System.exit(1);
		}
	}

	private static Properties loadProperties() throws IOException {
		Properties props = new Properties();
		File file = new File(RACKSPACE_PROPERTIES_FILE);
		if (!file.exists()) {
			props.put("user", "MY_RACKSPACE_USERNAME");
			props.put("apiKey", "MY_RACKSPACE_ACCESS_KEY");
			props.store(new FileWriter(file), "Generated properties file for ESM Demo Rackspace properties");
			throw new FileNotFoundException("Could not find properties file: " + file.getAbsolutePath()
					+ ". A default one was created, " + "please set the values according to your Rackspace account");
		}

		Reader reader = new FileReader(file);
		try {
			props.load(reader);
			return props;
		} finally {
			reader.close();
		}

	}

	private static void startGUI(final String locator) throws IOException {
		ProcessBuilder pb = new ProcessBuilder(new String[] { "cmd.exe", "/c", "start",
				"c:/Gigaspaces/gigaspaces-xap-premium-7.1.1-ga/bin/gs-ui.bat" });
		pb.environment().put("LOOKUPLOCATORS", locator);
		pb.start();

	}

	private static String startESMNode(final String name, final String user, final String apiKey) throws IOException {

		Deployer deployer = null;
		try {
			deployer = new Deployer(user, apiKey);
			Server server = null;

			// First check if a server already exists with this name
			System.out.println("Checking if server already exists.");
			server = deployer.getServer(name);
			if (server != null) {

				final String locator = (String) server.getAddresses().getPrivateAddresses().toArray()[0];
				System.out.println("Server " + name + " already exists, locator = " + locator);
				return locator;
			}

			// If not, the GigaSpaces Rackspace deployer creates a
			// Rackspace cloud server and sets up Gigaspaces on it.
			server = deployer.launchESMMachine(name, null);

			if (server == null) {
				throw new IllegalStateException("Failed to start GigaSpaces server!");
			}

			// Use the private address as a locator, since we are deploying
			// from inside the Rackspace cloud.
			final String locator = (String) server.getAddresses().getPrivateAddresses().toArray()[0];

			return locator;
		} finally {
			deployer.destroy();
		}

	}

	private ESMDemo() {

	}

}
